/********************************************************************************

 Programmed by SEUNG-JO BAE 

  - Goal: Prefix (Exclusive) and Reduction Sum

  - Algorithm: Kurmar's Book Program 3.6 & 3.7

  - Input: Integer number

  - Output: Integer number

  - Version 1.0 (2/13/95)

  - Communication: CMMD Swap

  - Note: Combine Prefix (Exclusive) Sum and Reduction Sum => PRS


********************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmsys/cm_fcntl.h>  /* fcntl( fileno(stdout), F_SETFL, O_APPEND); */

#include "global0.h"

#include "function0.h"


void initialize()
{
   Boolean error;
   int i;

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   ori_nproc = CMMD_partition_size();
   do {
      printf("\nEnter number of processors(Max=%d): ", ori_nproc);
      scanf("%d", &nproc);
      error = ( !Blog2(nproc) || nproc<2 || nproc>ori_nproc || nproc>N );
      if (error)
	 printf("\n??? Improper input. Try again!\n");
    } while (error);
    i = CMMD_reset_partition_size(nproc);
    if (i != nproc) 
       CMMD_error("\nERROR: CMMD Partition failed!\n");
   
   my_pnum  = CMMD_self_address(); 
 }



void PRS_dim0(my_num, d, p_result, r_result)
  int my_num, d, *p_result, *r_result;
{

  int out_mesg, in_mesg; 
  int partner;
  int i;

  *p_result = 0;      /* exclusive prefix */

  out_mesg = my_num;  /* same as r_result */

  CMMD_sync_with_nodes();

  for (i=0; i<d; ++i) {
    partner = my_pnum ^ pow2(i);

    CMMD_swap(partner, 
	      &in_mesg, 4,       /* receive */
	      &out_mesg, 4);     /* send */

    out_mesg += in_mesg;

    if (partner < my_pnum)
      *p_result += in_mesg;
  }

  *r_result = out_mesg;

  CMMD_sync_with_nodes();
}




void main()
{

  int my_num, d, p_result, r_result;

  initialize();

  my_num = 1;
  d = Ilog2(nproc);

  PRS_dim0(my_num, d, &p_result, &r_result);

  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);

  printf("\n [pn=%2d] in=%d, prefix=%d, reduction=%d \n", 
         my_pnum, my_num, p_result, r_result);
  fflush(stdout);


}
